NUM     [0-9]+
MACPAIR [[:xdigit:]][[:xdigit:]]
MACA    {MACPAIR}:{MACPAIR}:{MACPAIR}:{MACPAIR}:{MACPAIR}:{MACPAIR}
IPA     [0-9]+\.[0-9]+\.[0-9]+\.[0-9]+
IDENT   [[:alpha:]][[:alnum:]]*
SPACE   [ \t\n]+

%option noyywrap

%%

{ 
"if"     { return IF;    }
"else"   { return ELSE;  }
"+="     { return ADDEQ; }
"<="     { return LEQ;   }
">="     { return GEQ;   }
"=="     { return EQ;    }
"||"     { return OR;    }
"&&"     { return AND;   }
"!="     { return NEQ;   }
"<"      { return LE;    }
">"      { return GE;    }

{MACA}   { // Create a symbol with the MAC address and return it:
           mac_addr val;
	   if(!val.fromstr(yytext)) yyerror("Bad format of MAC address");
	   
	   yylval.nmacvar=*new macConst(val);
	   return MACVAR;
	 }
	   
{IPA}    { // Create a symbol with the IP address and return it:
           ip_addr val;
	   if(!val.fromstr(yytext)) yyerror("Bad format of IP address");
	   
	   yylval.nipvar=*new ipConst(val);
	   return IPVAR;
	 }
	   
{NUM}    { // Create a symbol with integr value and return it:
           int val=atoi(yytext);
	   yylval.nintvar=*new intConst(val);
	   return INTVAR;
	 }

{IDENT}  { // Lookup symbol in symbol tabel. Note that it is automatically
           // created if it doesn't exist.
           Symtable::SymEntry* se=symtabel.getSymbol(yytext);
	   
	   switch(se->type) {
	     case ParseProgram::tINT: yylval.nintvar=se->value.int_value; return INTVAR;
	     case ParseProgram::tMAC: yylval.nmacvar=se->value.mac_value; return MACVAR;
	     case ParseProgram::tIP:  yylval.nipvar=se->value.ip_value;   return IPVAR;
	   }
	 }

{SPACE}
.        { return yytext[0]; }
}