// This is the wipld daemon process.
//
// Possible improvements: 
//
// 1) Make the process fork later making it possible to write error messages 
//    to the console instead of only to

#include "common.h"
#include "parseprg.h"
#include "shmem.h"
#include "config.h"
#include "logfile.h"
#include "pcapintf.h"
#include "sigctl.h"

#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/param.h>
#include <errno.h>
#include<sys/stat.h>

#include <sys/ioctl.h>
#include <fcntl.h>

//-----------------------------------------------------------------------------

// This memory creates the shared memory area. When given packets to
// the consume function it then updates the area accordingly.

class CardCollect:ParseProgram {
  ShMemServer serv;
  LogFile* lf;
  
  int counterc;    
  int memsize;
  
  ServerMem* ca; // Set to the shared memory area by consume
  char* error;   // set by macIncr on error
  virtual void macIncr(mac_addr,int counterc, int** counters, int size);  
  virtual types getSymbol(char* name, void*& location);
  
  int ct;  // If true program contains a variabel reference that requires
           // localtime to be called for each packet.
  tm time;

  public:
  
  // This inits the daemon.
  // Returns 0 or an error message
  char* init(Config*, LogFile*);    
  
  // After successfull call to init this can be called:
  int getShMemSize() { return memsize; }
  
  PcapInterface::Packet pack;

  // Consumes the packet pack. Returns 0 on success or an error message on
  // a fatal error.
  char* Consume();
};

ParseProgram::types CardCollect::getSymbol(char* name, void*& location) {
  if(!strcmp(name,"srcmac"))  { location=&pack.maca_src; return tMAC; }
  if(!strcmp(name,"dstmac"))  { location=&pack.maca_dst; return tMAC; }
  if(!strcmp(name,"srcip"))   { location=&pack.ip_src; return tIP; }
  if(!strcmp(name,"dstip"))   { location=&pack.ip_dst; return tIP; }
  if(!strcmp(name,"srcport")) { location=&pack.ipport_src; return tINT; }
  if(!strcmp(name,"dstport")) { location=&pack.ipport_dst; return tINT; }
  if(!strcmp(name,"size"))    { location=&pack.ip_len; return tINT; }
 
  if(!strcmp(name,"hour"))    { location=&time.tm_hour; ct=1; return tINT; }
  if(!strcmp(name,"min"))     { location=&time.tm_min;  ct=1; return tINT; }
  if(!strcmp(name,"wday"))    { location=&time.tm_wday; ct=1; return tINT; }
  if(!strcmp(name,"mday"))    { location=&time.tm_mday; ct=1; return tINT; }
  
  // Else set the type to integer and the let the variabel be initialized
  // automatically:
  return tINT;
}  

char* CardCollect::init(Config* cnf, LogFile* l) {
  counterc=cnf->counterc();
  lf=l;
  int maxcardc=cnf->maxcardc();
    
  // Check parameters:
  if(maxcardc<=0) return "At least one card should be allowed";
  if(counterc<=0 || counterc>=maxcounterc) {
    snprintf(errmsg,sizeof(errmsg),"Between 1 and %i counters must be defined",maxcounterc); return errmsg;
  }    
  
  // Compile program and return error message on error:
  ct=0;
  char* c;
  if(c=setProgram(cnf->program())) return c;
		      
  // Create the shared memory:
  void* memory;
  memsize=sizeof(ServerMem)+maxcardc*sizeof(CardInfo);
  c=serv.init(cnf->daemonfile(),memsize,memory);
  if(c) return c;
  
  // An initialize it:
  ServerMem* ca=(ServerMem*)memory;    
  
  ca->Counterc=counterc;
  for(int a=0; a<counterc; a++) {
    strncpy(ca->CounterNames[a],cnf->countername(a),MaxCounterNameLen);
    ca->CounterNames[a][MaxCounterNameLen]='\x0';
  }
  for(int a=counterc; a<maxcounterc; a++) {
    strcpy(ca->CounterNames[a],"(unused)");
  }
  
  ca->maxcardc=maxcardc;
  ca->curcardc=0;
  
  gettimeofday(&ca->start.tv,0); 
  serv.UnLock(ca);
  
  return 0;
}

char* CardCollect::Consume() {
  // Lock the shared memory:
  ca=(ServerMem*)serv.Lock();
  if(ca==0) return LockErrString;
  
  // Calculate the time value if used by the program:
  if(ct) time=*localtime(&pack.time.tv.tv_sec);
  
  // Execute the program
  error=0;
  execProgram();
  
  // And unlock the memory:
  serv.UnLock(ca);
  
  if(error) return error;  
  return 0;
}  

void CardCollect::macIncr(mac_addr ma, int counterc, int** counters, int size) {    
  // Try find the card in our table:
  CardInfo c; c.macaddr=ma; 
  CardInfo* a=(CardInfo*)bsearch(&c,&ca->cards,ca->curcardc,sizeof(CardInfo),CardInfo::compare_mac);

  if(a==0) {
    // The card could not be found - insert it:
    if(ca->curcardc>=ca->maxcardc) {
      snprintf(errmsg,sizeof(errmsg),"More than %i cards in table\n",ca->maxcardc);
      error=errmsg;
      return;
    }
    
    // Find index to insert card at:
    int i;
    for(i=0; i<ca->curcardc; i++)
      if(mac_addr::compare(ca->cards[i].macaddr,ma)>=0) break;
    a=&ca->cards[i];
    
    // Insert the card at index i - move the other cards:
    int b;
    for(b=ca->curcardc; b>i; b--) ca->cards[b]=ca->cards[b-1];
    ca->curcardc++;
    
    // And insert the new card (make non-used counters zero, too)
    a->macaddr=ma;
    a->ipaddr.setNull();
    for(int b=0; b<maxcounterc; b++) a->Counters[b]=0;    
  }
  
  // Update counter information for the card. Note that we don't add
  // negative values since this will confuse som clients.
  for(int b=0; b<counterc; b++) {
    int cnt=*counters[b];
    if(cnt>=0 && cnt<this->counterc) // If valid counter:
      a->Counters[cnt]+=size<0 ? -size : size;
  }
  
  // Update information about last update:
  gettimeofday(&a->updated.tv,0);  
  
  // Update IP address if packet was to or from this card. We do it here
  // and not in Consume so we don't have to search for a card several times.
  ip_addr ip;
  if(ma==pack.maca_src) 
    ip=pack.ip_src;
  else if(ma==pack.maca_dst)
    ip=pack.ip_dst;
  else return;
        
  if(ip!=a->ipaddr) {
    char buf[40];
    snprintf(buf,40,"%s",ip.tostr());
    lf->log(LogFile::ipchange,
      "IP address for card %s changed from %s to %s",
      ma.tostr(),a->ipaddr.tostr(),buf);
    a->ipaddr=ip;
  }
}

//-----------------------------------------------------------------------------

// Class used to create file with the process id of the daemon.
class PidFile {
  char* file;

  static void programAbort(void *p) {
    // Delete file:
    unlink(((PidFile*)p)->file);
  };
  
  public:
  PidFile()  : file(0) { registerSignalHandler(this,programAbort,signalTerm); }
  ~PidFile() { unregisterSignalHandler(this,programAbort); programAbort(this); delete[] file; }
  
  char* init(char* filename) {
    blockSignals();
    file=new char[strlen(filename)+1];
    strcpy(file,filename);
    unblockSignals();
    int h=open(filename,O_CREAT|O_WRONLY|O_APPEND,
       S_IRUSR|S_IWUSR|
       S_IRGRP|S_IWGRP|
       S_IROTH);
    if(h==-1) {    
      snprintf(errmsg,sizeof(errmsg),"Could not create pid file %s: %s",filename,strerror(errno));
      return errmsg;
    }
    char buf[40];
    snprintf(buf,sizeof(buf),"%i\n",(int)getpid());
    write(h,buf,strlen(buf));
    close(h);
    return 0;
  }
};

main(int argc, char*argv[]) {  
  // Find name of configuration file:
  char* conffile;
  if(argc==1)
    conffile=DEFAULT_CNFFILE;
  else if(argc==2)
    conffile=argv[1];
  else {
    fprintf(stderr,"Syntax: %s [conffile]\n",argv[0]);
    return 1;
  }
  
  // Read configuration file:
  Config* cnf=new Config();
  char* c=cnf->readConfFile(conffile);
  if(c) {
    fprintf(stderr,"%s\n",c);
    return 2;
  }
  
  // Open log file:
  LogFile lf;
  c=lf.init(cnf);
  if(c) {
    fprintf(stderr,"%s\n",c);
    return 2;
  }
  
  // We have now opened the logfile. Because of this we are able to write
  // errormssages to the log file instead of to the terminal. We can there-
  // fork and let the parent process exit. On the other hand we can't fork
  // later because of problems with destructor calls and other things.
  int i=fork();
  if(i==-1) { // On error:
    fprintf(stderr,"Could not fork: %s\n",strerror(errno));
    return 3;
  }
  if(i!=0) { // On sucess and parent process
    return 0;
  }
  
  // We are now the child process.
  
  // Change to root directory to allow unmount of cwd:
  chdir("/");
  
  // Disassociate from process group so we don't receive signals for the group:
  setpgrp();
  
  // Disassociate from controling terminal:
  int h;
  if(h=open("/dev/tty",O_RDWR)>=0) {
    ioctl(h,TIOCNOTTY,(char*)0);
    close(h);
  }
  
  // Close all filehandels except for the logfile:
  lf.closeAllOtherFiles();
  
  // Initialize daemon and compile program:
  CardCollect cc;
  c=cc.init(cnf,&lf);

  // Create pid file.
  PidFile pf;
  c || (c=pf.init(cnf->pidfile()));
  
  // Start reading from device:
  PcapInterface pi;  
  c || (c=pi.init(cnf->filter(),cnf->netdev(),&lf));
  
  lf.log(LogFile::status,"---");

  if(c) {
    lf.log(LogFile::status,"Daemon could not start: %s",c);
    return 1;
  }
  
  // Print status information:
  lf.log(LogFile::status,"Daemon initialized with pid %i and shared memory size %i",getpid(),cc.getShMemSize());
  lf.log(LogFile::status,"Device: %s, Logfile: %s, Pidfile: %s, Daemonfile: %s",pi.getDev(),cnf->logfile(),cnf->pidfile(),cnf->daemonfile());
  
  delete cnf; // Save a bit memory...
    
  for(;;) { 
    c=pi.nextPacket(cc.pack);
    c || (c=cc.Consume());
    if(c) {
      lf.log(LogFile::status,"Daemon terminated: %s",c);
      return 1;
    }
  }
}  
